/* $Id: check_ps.c,v 1.5 2000/10/10 22:10:43 dps Exp $ */
/* The main detection code. */
/* Just runs this the background and detects nasty version of ps */
/* (c) D.P.Simpson 1997, rediristibution and use encouraged, provided */
/* you comply with the GPL. */

#include "config.h"
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif /* HAVE_CTYPE_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */
#include "prog.h"
#include "cfg_check.h"
#include "numbuf.h"
#include "libdetail/detail.h"

#ifndef HAVE_LSTAT
#define lstat(x,y) stat(x,y)
#endif /* HAVE_LSTAT */

/* Test a process id */
static int process_ok(numbuf *pids, int pid, const numbuf *fpl, numbuf *nl,
		      int (*current_proc)(int))
{
    int i;

    for (i=0; i<pids->pos; i++)
    {
	if (pids->nums[i]==pid)
	{
	    pids->nums[i]=-2;	/* Used, OK */
	    return 1;		/* I am happy */
	}
    }

    if (!(*current_proc)(pid))
    {
	/* Check the fake pid list */
	for (i=0; i<fpl->pos; i++)
	{
	    if (fpl->nums[i]==pid)
	    {
#ifdef DEBUG
		fprintf(stderr, "Hit...fake pid %d\n", pid);
#endif		
		log("Fake pid %d detected", pid);
	    }
	}
	/* Add to list of fake pids */
	if (nl->pos==nl->max && expand_buf(nl)==0)
	{
	    log("Fake pid list overflow, size %d", nl->pos);
	}
	nl->nums[(nl->pos)++]=pid;
	return 2;		/* Already dead. ps probably tickles this. */
    }
    return 1;
}


/* ps output analyser, mainly just toilet roll. Some anti-DoS stuff too
 * lest jokers write a DoS against this program version of ps. However
 * that would be obviously hacked so they might not take the risk. 
 */
static int run_ps(numbuf *pids, const numbuf *fpl, numbuf *nl,
		  int (*current_proc)(int))
{
    FILE *in;
    int c, num, nc, i;
    static const char *const ps_args[]={"ps", quote(PS_ARGS), NULL}; 
    enum {SKIP_WS, GET_NUM, SKIP_END, DOS_ATTACK} line;
    int retest=0;

    if ((in=safe_spawn(quote(PS_PATH), ps_args, PS_MAXTIME))==NULL)
	return 0;

    line=SKIP_END;		/* Skip 1st line */
    nc=0;			/* character count is zero */
    while(line!=DOS_ATTACK && (c=fgetc(in))!=EOF)
    {
       switch(line)
       {
       case SKIP_WS:
           if (isspace(c))
	   {
	       nc++;
	       if (nc>TOLERANCE) /* Test for DoS attack... */
	       {
		   log("Attempted DoS attack: >" quote(TOLERANCE)
		       " bytes leading white space");
		   line=DOS_ATTACK;
		   break;
	       }
               continue;
	   }
           num=0;
	   nc=0;
           line=GET_NUM;
           /* FALL THRU */

       case GET_NUM:
           if (isdigit(c))
           {
	       nc++;
	       if (nc>NUM_TOLERANCE)
	       {
		   log("Attempted  DoS attack: >" quote(NUM_TOLERANCE)
		       " digits in pid");
		   line=DOS_ATTACK;
		   break;
	       }
               num=num*10+(c-'0');
               continue;
           }
           if (nc==0)
	   {
	       log("Attempt to attack me by not starting with a number");
	   }
	   else
	   {
	       switch(nc=process_ok(pids, num, fpl, nl, current_proc))
	       {
	       case 0:
		   nuke_proc(num); /* Ask no questions, nuke */
		   retest=1;
		   break;
		   
	       case 1:
		   break;
		   
	       case 2:
#ifdef DEBUG
		   fprintf(stderr, "retest %d\n", num);
#endif
		   retest=1;
		   break;
		   
	       default:
		   log("process ok returned %d", nc);
		   break;
	       }
	   }
	   nc=0;
	   line=SKIP_END;
	       /* FALL THRU */

       case SKIP_END:
	   nc++;
	   if (nc>TOLERANCE) /* Test for DoS attack... */
	   {
	       log("Attempted DoS attack: more than " quote(TOLERANCE)
		   " before newline");
	       line=DOS_ATTACK;
	       break;
	   }
           if (c=='\n')
               line=SKIP_WS;
           break;
	       
       case DOS_ATTACK:
	   log("Killing evil version of ps", spawned_pid());
	   kill(spawned_pid(), SIGKILL); /* Kill evil ps */
	   break;

       default:
           log("Imposible state %d", line);
           fclose(in);
           return 0;
       }
    }
    kill(spawned_pid(), SIGKILL);		/* Kill ps (stop ps DoS attack) */
    /* No need to wait or turn off alarm, SIGCHLD handler does that */

    /* Did ps tell the truth or was there something it was not telling us?
     * current_process is a function to eliminate false positives due to
     * races (a process get found by get_pids but exits before we execute
     * ps). A small risk of a false -ve is introduced... */
    for (i=0; i<pids->pos; i++)
    {
	if (pids->nums[i]!=-2 && (*current_proc)(pids->nums[i]))
	{
	    if (nuke_proc(pids->nums[i]))
		    log("Hidden pid %d", pids->nums[i]);
	}
    }
    /*    log_flush();		/* Flush log lines. */
    return (retest==0) ? 1 : 2;
}



/* demon body */
/* The demon uses *two* fake pid lists to avoid generating alarms due to race
 * conditions. If a fake pid turns up twice then it is not a race but somethig
 * rather nastier (thanks to bugtraq for the race and NASA for reminding me
 * that fake versioons of ps might add some things that should be around but
 * are not actually present). */
void demon_body(numbuf *(*pid_list)(void),  int (*current_proc)(int))
{
    int idx,r,did_check;
    numbuf *alrm[2], *nums;

    ignore_signals();
    alrm[0]=allocate();
    alrm[1]=allocate();
    idx=0;
    did_check=0;		/* Avoid spamming admin */
    while(1)
    {
       log_start();		/* Start recording period */
       if ((nums=(*pid_list)())==NULL)
       {
           log("pid_list failed");
	   log_flush();		/* Send logged lines to log */
	   check_time(0, MAXDIFF); /* Check clock */
           sleep (SHORTNAP);	/* Retry after a pause */
	   check_time(SHORTNAP, SHORTNAP+TOL_TIME); /* Check clock */
           continue;
       }
       r=run_ps(nums, alrm[idx], alrm[1-idx], current_proc); /* Runs ps */
       if (r==0)
       {
	   check_time(0, MAXDIFF); /* Check clock */
	   sleep(SHORTNAP);
	   check_time(SHORTNAP, SHORTNAP+TOL_TIME); /* Check clock */	   
	   continue;		/* Retry after a pause */
       }
       empty_nums(alrm[idx]);	/* Delete fake pid list */
       idx=1-idx;		/* Switch to bew fake pid list */
       free_nums(nums);		/* Free memory */
       log_flush();		/* Send logged liens to log */
       if (r==1 && did_check==0)
       {
	   did_check=1;		/* Set flag */
	   check_time(0, MAXDIFF); /* Check clock */
	   sleep(RECHECKNAP);	/* Sleep a bit */
	   check_time(RECHECKNAP, RECHECKNAP+TOL_TIME); /* Check clock */
	   continue;		/* Check ps again */
       }
       did_check=0;		/* Did check */
       check_time(0, MAXDIFF);	/* Check clock */
       sleep(LONGNAP);		/* Sleep alot */
       check_time(LONGNAP, LONGNAP+TOL_TIME);	/* Check clock */
    }
}

/* One shot verson of demon body */
int one_shot(numbuf *(*pid_list)(void), int (*current_proc)(int))
{
    int r, idx;
    numbuf *alrm[2], *nums;

    ignore_signals();
#ifdef DEBUG
    fputs("entered one_shot\n", stderr);
#endif
    system_ok_msg=0;		/* Message afer this are provlems */
    alrm[0]=allocate();
    alrm[1]=allocate();
    idx=0;
    while(1)
    {
	if ((nums=(*pid_list)())==NULL)
	{
	    log("pid_list failed");
            /*    log_flush();	/* Send logged lines to log */
	    check_time(0, MAXDIFF); /* Check clock */
	    sleep (SHORTNAP);	/* Retry after a pause */
	    check_time(SHORTNAP, SHORTNAP+TOL_TIME);
	    continue;
}
	r=run_ps(nums, alrm[idx], alrm[1-idx], current_proc); /* Runs ps */
#ifdef DEBUG
	fprintf(stderr, "exit with r=%d\n", r);
#endif
	if (r==0)
	{
	    log("runps failed");
	    check_time(0, MAXDIFF); /* Check clock */
	    sleep(SHORTNAP);
	    check_time(SHORTNAP, SHORTNAP+TOL_TIME); /* Check clock */
	    continue;		/* Retry after a pause */
	}
	free_nums(nums);	/* Free memory */
	if (r==1 || idx>0) 
        {
            log_flush();	/* Send logged lines to log */
	    return (idx==0) ? 0 : 1;
        }
	if (r!=1)
	{
	    log("Some pids seem to be fake");
	    idx=1-idx;
	    check_time(0, MAXDIFF); /* Check clock */
	    sleep(RECHECKNAP);
	    check_time(RECHECKNAP, RECHECKNAP+TOL_TIME); /* Check clock */
#ifdef TEST
	    printf("Please press enter to do retest (^Z enabled now)\n");
	    fflush(stdout);
	    signal(SIGTSTP, SIG_DFL);
	    fgetc(stdin);
	    signal(SIGTSTP, SIG_IGN);
	    printf("Thankyou (^Z now disabled again)\n");
	    fix_time();
#endif /* TEST */
	    log("rechecking ps");
	    continue;
	}
    }
    fputs("Oops! Got to IP2\n", stderr);
    return 2;
}
