/* $Id: tools.h,v 1.2 2000/05/14 19:16:26 dps Exp $ */
/* Tools for paraniod reading of files, etc */

#ifndef __CPS_DETAIL_H__
#define __CPS_DETAIL_H__ 1

/* checkps' config.h is protected against multiple inclusion so this
 * is safe. Standard config.h is not protected in current autoconf */
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif /* HAVE_STDIO_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */

/* Limits */
#ifndef HEX_MAXDIGIT
#define HEX_MAXDIGIT 20
#endif /* HEX_MAXDIGIT */

#ifndef DEC_MAXDIGIT
#define DEC_MAXDIGIT 20
#endif /* DEC_MAXDIGIT */

#ifndef MAX_WS
#define MAX_WS 200
#endif /* WS_MAX */

#ifndef MAX_LINE
#define MAX_LINE 1024
#endif /* MAXL_LINE */

#ifndef MAX_STRING
#define MAX_STRING 100
#endif /* MAX_STRING */

/* White space */
extern int absorb_ws(FILE *);		       /* Safe white space absorbition  */
extern int check_eol(FILE *);		       /* Check <optional ws>\n */

/* Numbers */
extern int readdec(FILE *, int *);	       /* Safe read decimal number */
extern unsigned int readhex(FILE *, int *);    /* Read hex width digit count */

/* string.c */
extern int get_string(FILE *, char *, int);	 /* Read string */
extern int skip_string(FILE *, int);		 /* Skip string */
extern int require_string(FILE *, const char *); /* Check fixed string */

/* Nice front end to above */
extern int check_fmt(FILE *, const unsigned char *); /* printf style front end to above */

/* Line eater */
extern int eat_line(FILE *in);		       /* Safe eat line */

/* Look for a list of strings */
extern int check_list(FILE *, const char *const *, int);

/* Paranoid file open */
extern FILE *open_proc_obj(const char *fname, const int);
#define open_procfile(x) open_proc_obj(x, S_IFREG)
#define open_procdir(x) open_proc_obj(x, S_IFDIR)
#define open_procsock(x) open_proc_obj(x, S_IFSOCK)

/* IP address scanner (ipread.c) */
extern int read_ip(FILE *, struct sockaddr_in *);

/* safe_exec.c */
extern int safe_exec(const char *prog, const char *const *argv, int fd);

/* catch_child.c */
extern FILE *safe_spawn(const char *const, const char *const *, unsigned int);
extern pid_t spawned_pid(void);

/* inode-cache.c */
extern void inode_cache(const char *fname, ino_t inde, dev_t dev, time_t mtime,
			int fixed);
extern void icache_init(void);
#define INODE_KEEP 1		/* Do not allow LRU to remove this entry  */
#define INODE_MTIME 2		/* Monitor mtime for this inode */

#endif /* __CPS_DETAIL_H__ */
