/* $Id: fd_analysis.c,v 1.2 2000/10/10 22:10:44 dps Exp $ */
/* filedescriptor analysis for linux */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif /* HAVE_SYS_FCNTL_H */
#include "linux/os-cfg.h"
#include "linux/proc.h"
#include "libdetail/detail.h"
#include "prog.h"

/* fs type support */
#ifdef STATFS_WORKS
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif /* HAVE_SYS_VFS_H */
#endif /* STATFS_WORKS */
 	
int fd_ok(int pid)
{
    char filename[FNAME_LEN];	/* File name */
    struct stat stbuf;
    int errno_save;
    int fd;
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    struct statfs fts;
#endif

    if (snprintf(filename, FNAME_LEN, "/proc/%d/fd", pid)>=FNAME_LEN)
	return 0;

    /* Try to open the file */
    fd=open(filename, O_RDONLY);
    if (fd==-1)
    {
	errno_save=errno;	/* Save errno */
	kill(pid, 0);		/* Check if process still exists */
	if (errno!=ESRCH)
	{
	    log("pid %d: could not open /proc/%d/fd (error %s)", pid, pid,
		strerror(errno_save));
	}
	return 0;
    }

    /* Check the fd refers to a directory */
    if (fstat(fd, &stbuf)<0)
    {
	log("pid %d: could not stat fd %d (errno %s)", pid, fd, strerror(errno));
	close(fd);
	return 0;
    }
    if (!S_ISDIR(stbuf.st_mode))
    {
	log("pid %d: /proc/%d/fd not a directory", pid, pid);
	close(fd);
	return 0;
    }
    
    /* Check the fs type */
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    if (fstatfs(fd, &fts)!=0)
    {
	log("pid %d: Unexpected failure of statfs (fd %d, error %s)",
	    pid, fd, strerror(errno));
	close(fd);
	return 0;
    }
    if (fts.f_type!=CPS_PROC_FS_TYPE)
    {
	log("pid %d: Uh. Uh. Detected hacking attempt: wrong fs type!", pid);
	close(fd);
	return 0;
    }
#endif /* HAVE_STATFS and STATFS_FST and STATFS_WORKS */
    close(fd);
    return 1;
}


void fd_analysis(int pid)
{
    char filename[FNAME_LEN];	/* File name */
    struct stat stbuf;
    int i, fd;

    if (pid==0)
	return;

    if (!fd_ok(pid))
	return;

    for (i=0; i<MAX_FD; i++)
    {
	if (snprintf(filename, FNAME_LEN, "/proc/%d/fd/%d", pid, i)>=FNAME_LEN)
	    continue;

	if ((fd=open(filename, O_RDONLY))<0)
	    continue;
	if (fstat(fd, &stbuf)<0)
	{
	    log("pid %d: fstat failed for fd %d (error %s)", pid, i,
		strerror(errno));
	    continue;
	}

	if (S_ISSOCK(stbuf.st_mode))
	{
	    log_socket(fd, pid, &stbuf);
 	}
	else if (S_ISREG(stbuf.st_mode))
	{
	    log("pid %d: fd %d is regular file dev %x (%d:%d), inode %x (owner u=%d g=%d)",
		pid, i,
		(int) stbuf.st_dev, (int) (stbuf.st_dev>>8), (int) (stbuf.st_dev & ((1<<8)-1)),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	}
	else if (S_ISDIR(stbuf.st_mode))
	{
	    log("pid %d: fd %d is directory dev %x (%d:%d), inode %x (owner u=%d g=%d)",
		pid, i, 
		(int) stbuf.st_dev, (int) (stbuf.st_dev>>8), (int) (stbuf.st_dev & ((1<<8)-1)),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	}
	else if (S_ISFIFO(stbuf.st_mode))
	    log("pid %d: fd %d is fifo dev %x (%d:%d) inode %x (owner u=%d g=%d)", pid, i,
		(int) stbuf.st_dev,(int) (stbuf.st_dev>>8), (int) (stbuf.st_dev & ((1<<8)-1)),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	else if (S_ISCHR(stbuf.st_mode))
	{
	    log("pid %d: fd %d is block dev %x (%d:%d)"
		" [file: dev %x (%d:%d) inode %x (owner u=%d g=%d)]", pid, i,
		(int) stbuf.st_rdev, (int) (stbuf.st_rdev>>8), (int) stbuf.st_rdev & ((1<<8)-1),
		(int) stbuf.st_dev, (int) (stbuf.st_dev>>8), (int) stbuf.st_dev & ((1<<8)-1),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	}
	else if (S_ISBLK(stbuf.st_mode))
	    log("pid %d: fd %d is block dev %x (%d:%d)"
		" [file: dev %x (%d:%d) inode %x (owner u=%d g=%d)]", pid, i,
		(int) stbuf.st_rdev, (int) (stbuf.st_rdev>>8), (int) (stbuf.st_rdev & ((1<<8)-1)),
		(int) stbuf.st_dev, (int) (stbuf.st_dev>>8), (int) (stbuf.st_dev & ((1<<8)-1)),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	else
	    /* Hopefully impossible---I think the tests above are exhaustive */
	    log("pid %d: fd %d is open and unknwon type dev %x (%d:%d) inode %x"
		" (owner u=%d g=%d)", pid, i,
		(int) stbuf.st_dev, (int) (stbuf.st_dev>>8), (int) (stbuf.st_dev & ((1<<8)-1)),
		(int) stbuf.st_ino, (int) stbuf.st_uid, (int) stbuf.st_gid);
	close(fd);
    }
}
