/* $Id$ */
/* Log suspect netstat evidence */

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif /* HAVE_STDIO_H */
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */#
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */

#include "detail.h"
#include "socklist.h"

static socklist *filter_out(socklist *s, const int mask, const int v)
{
    socklist *scan, *list_top;

    list_top=s;
    scan=s;
    scanp=&list_top;
    while (scan!=NULL)
    {
	if (((scan->matchval) & mask)==v)
	{
	    /* Socket appeared in all lists, kill it off */
	    *scanp=scan->next;
	    free_sock(scan);
	    scan=*scanp;
	    continue;
	}
	/* Socket not listed everywhere, keep it */
	scanp=&(scan->next);
	scan=scan->next;
    }
    return list_top;
};

static socklist *badlist(numbuf *pids, int (*current_proc)(int))
{
    socklist *pid, *netstat, *proc;
    socklist **scanp, *scan;

    proc=sockl_proc();			 /* Get socket list from /proc */
    netstat=run_netstat();		 /* Get socket list from netstat command */
    pid=pid_scan(pids, current_proc);	 /* Get socket list by scanning processes */
    socklist_setval(proc, 1);		 /* Mark proc list by bit 0 */
    socklist_setval(netstat, 2);	 /* Mark netstat output by bit 1 */
    socklist_setval(pid, 4);		 /* Mark process id list by bit 2 */

    proc=socklist_aggregate(netstat, proc);  /* Merge proc and netstat */
    pid=socklist_aggregate(proc, pid);	     /* Merge pid and (proc and netstat) */

    pid=filter_out(pid, 7, 7);		 /* Kill items on all 3 lists */
    return pid;
}

static void log_socklist(const socklist *l)
{
    struct sockaddr_in *s, *d;
    struct sockaddr_un *u;
    const __sock_ref *ref;
    const char *src, *dst;
    char *tbuf;

    while(l!=NULL)
    {
	switch(l->fam)
	{
	case AF_INET:
	    s=(struct sockaddr_in *) l->src;
	    d=(struct sockaddr_in *) l->dest;

	    /* Get source and destination addresses, complexified by
	       fact later inet_ntoa clobbering previous ones */
	    dst=inet_ntoa(s->sin_addr);
	    if ((tbuf=malloc(strlen(dst)+1))==NULL)
		src="<malloc failure>";
	    else
	    {
		strcpy(tbuf, dst);
		src=tbuf;
	    }
	    dst=inet_ntoa(d->sin_addr);
	    log("Internet socket type %d from %s:%d to %s:%d inode %d, user %d",
		(int) l->type, src, (int) s->sin_port, dst, (int) d->sin_port,
		(int) l->inode, (int) l->uid);
	    for (ref=d->refs; ref!=NULL; ref=ref->next)
	    {
		log("socket type %d from %s:%d to %s:%d is fd %d in process %d\n",
		    (int) l->type, src, (int) s->sin_port, dst, (int) d->sin_port,
		    ref->fd, (int) ref->pid);
	    }
	    break;

	case AF_UNIX:
	    u=(struct sockaddr_un *) l->src;
	    log("UNIX socket type %d inode %d, path %s\n",
		(int) l->type, (int) l->inode, &(u->sun_path));
	    for (ref=d->refs; ref!=NULL; ref=ref->next)
	    {
		log("UNIX socket type %d inode %d is fd %d in process %d\n",
		    (int) l->type, (int) l->inode, ref->fd, (int) ref->pid);
	    }
	    break;
	   
	default:
	    log("Other socket type %d\n", (int) l->fam);
	    break;
	}
	l=l->next;
    }
}

void check_netstat(numbuf *pids, int (*current_proc)(int))
{
    socklist *a, *b;

    a=badlist(void);			 /* Get list of suspect sockets */
    if (a==NULL)
	return;				 /* Return happily if none */
    sleep(1);				 /* Sleep 1 second */
    b=badlist(void);			 /* Get list again */
    socklist_setval(a, 8);		 /* Use bit 3 for old list */
    b=socklist_aggregate(a, b);		 /* Merge a and b */
    b=filter_out(b, 8, 0);		 /* Filter out everything with bit 3 not set */
    if (b==NULL)
	return;				 /* Return happily, just freak timing */

    /* If something left here then it showed up twice, so complain loadly */
    log("rootkit netstat detected");	 /* log bad news about netscape */
    log_socklist(b);			 /* Log detials */
    socklist_zap(b);			 /* Free memory */
    return;				 /* Unhappy about this */
}
