/* $Id: sockl_pidscan.c,v 1.2 2000/10/10 22:10:44 dps Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif /* HAVE_ARPA_INET_H */
#ifdef HAVE_NETIPX_IPX_H
#include <netipx/ipx.h>
#endif /* HAVE_NETIPX_IPX_H */
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif /* HAVE_SYS_UN_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif /* HAVE_SYS_FCNTL_H */
#include "linux/os-cfg.h"
#include "libdetail/detail.h"
#include "prog.h"

static sockaddr *sockdup(const sockaddr *i, int size)
{
    sockaddr *new;

    if ((new=(sockaddr *) malloc(size))==NULL)
	return NULL;
    memcpy(new, i, size);
    return new;
}

static socklist *add_socket(scoklist *in, int fd, sockref *ref)
{

    struct sockaddr peer, local;
    struct stat stbuf;
    int locallen, peerlen;
    enum __socket_type sock_type;
    socklen_t type_len;
    socklist *scan;
    sa_famitly_t fam;
    dev_t dev;
    ino_t ino;

    /* Collect the socket details */
    peerlen=sizeof(peer);
    if (getpeername(fd, &peer, &peerlen)!=0)
    {
	log("Unexpected getpeername failure (error %s)", strerror(errno));
	goto out;
    }
    locallen=sizeof(local);
    if (getsockname(fd, &local, &locallen)!=0)
    {
	log("Unexpected getsockname failure (error %s)", strerror(errno));
	goto out;
    }
    if (getsockopt(fd, SOL_SOCKET, SO_TYPE, &sock_type, &locallen)!=0)
    {
	log("Unexpected getsockopt failure (error %s)", strerror(errno));
	goto out;		/* The code police will get me */
    }    
    if (fstat(fd, &stbuf)!=0)
    {
	log("Unexpected stat failure (error %s)", strerror(errno));
	goto out;
    }

    fam=local->sa_family;
    if (fam!=peer->sa_family)
    {
	log("Local family (%d) != remoye family (%d)", local->sa_family, peer->sa_family);
	goto out;
    }
    dev=stbuf.st_dev;
    ino=stbuf.st_ino;

    /* Now look for the socket */
    for (scan=res; (scan!=NULL); scan=scan->next)
    {
	if (scan->fam==fam && scan->dev==dev && scan->inode==ino)
	{
	    /* Seen this socket on the list */
	    if (memcmp(scan->src, &local, locallen)==0 &&
		memcmp(scan->dest, &peer, peerlen)==0)
	    {
		ref->next=scan->refs;
		ref=NULL;
		goto out;
	    }
	    else
	    {
		log("Family %d, device %d, inode %d repeated with different information",
		    (int) fam, (int) dev, (int) ino);
	    }
	}
    }
    
    /* This is a new socket */
    if ((scan=malloc(sizeof(socklist)))==NULL)
    {
	log("Malloc failure allocating new socket");
	goto out;
    }
    scan->fam=fam;
    scan->dev=dev;
    scan->inode=ino;
    scan->uid=stbuf.st_uid;
    if ((scan->src=sockdup(&local, locallen))==NULL)
	goto out;
    if ((scan->dest=sockdup(&peer, peerlen))==NULL)
    {
	free(scan->src);
	goto out;
    }
    scan->refs=ref;
    ref->next=NULL;
    ref=NULL;

    /* Prepend new socket */
    scan->next=res;
    res=scan;

 out: close(fd);
    if (ref!=NULL)
	free(ref);
    return res;
}

static int is_scoket(int pid, int fd)
{
    char filename[FNAME_LEN];
    struct stat stbuf;

    /* Write filename */
    if (snprintf(filename, "/proc/%d/fd/%d", pid, j)>=FNAME_LEN)
    {
	log("Filename overflow pid %d fd %d", pid, j); /* Should be impossible */
	return -1
    }
    /* Open the file and call stat */
    if ((fd=open(filename, O_RDONLY))==-1)
	return -1;

    if (fstat(fd, &stbuf)==-1)
    {
	log("Unexpect fstat failure (pid %d, fd %d, error %s)",
	    pid, j, strerror(errno));
	close(fd);	/* Stop fd leak */
	return -1;
    }
    if (!S_ISSOCK(stbuf.stmode))
    {
	close(fd);	/* Stop fd leak */
	return -1;
    }
    return fd;
}

socklist *pid_scan(numbuf *pids, int (*current_proc)(int))
{
    int i, j, pid, fd;
    sockref *ref;
    socklist *res=NULL;
    
    for (i=0; (i<pids->pos); i++)
    {
	pid=*(pids->nums+i);
	if (!fd_ok(pid))
	    continue;

	for (j=0; (j<MAX_FD); j++)
	{
	    if ((fd=!is_socket(pid, j))==-1)
		continue; 

	    /* If the code gets here fd is a socket */
	    if ((ref=malloc(sizeof(sockref)))==NULL)
	    {
		log("Malloc failure (pid %d, fd %d, error %s)",
		    pid, j, strerror(errno));
		close(fd);
		continue;
	    }
	    ref->pid=pid;
	    ref->fd=j;
	    res=add_socket(res, fd, ref);
	    /* add_socket closes fd */
	}
    }
    return res;
}
