/* $Id: log.c,v 1.5 2000/10/10 22:10:43 dps Exp $ */
/* Log stuff, common to all logging */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /*  HAVE_CONFIG_H */
#ifdef HAVE_CONFIG_H
#include <stdio.h>
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STDDEF_H
#include <stddef.h>
#endif /* HAVE_STDDEF_H */
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif /* HAVE_SYSLOG_H */
#ifdef HAVE_TIME_H
#include <time.h>
#endif /* HAVE_TIME_H */
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif /* HAVE_STDARG_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#include "cfg_log.h"
#include "cfg_smtp.h"
#include "prog.h"

/* Do **NOT** delete this test. It is here to ensure that you set the
 * proper details in cfg_smtp.h. This program is a security item and
 * getting the configuration and code right is CRITICAL, not optional! */
#ifndef RECIPIENTS
#error No recipients for E-mail defined.
#error Before building you **MUST** edit cfg_smtp.h. Do it now.
#endif /* RECIPIENTS */


/* File to send out to */
extern FILE *log_file;		/* From prog.c */

/* Buffer stats */
static char *buf=NULL;		/* Buffer */
static int used;		/* Bytes used */
static int bsize;		/* Total size in bytes */
static const char *const recipients[]=
{
    RECIPIENTS, NULL
};


/* Send the stuff to syslog (mail failed, so little choice) */
static void use_syslog(char *buf)
{
    char *scan, *ls;

    scan=ls=buf;
    while (*scan)
    {
	if (*scan=='\n')
	{
	    *scan='\0';
	    syslog(LOG_WARNING, "%s", ls);
	    ls=scan+1;
	}
	scan++;
    }
}

void log(const char *fmt, ...)
{
    char line[MAXLINE], *nb;
    int nl, newbuf, bs;
    struct tm *time_p;
    time_t tim;
    size_t tim_sz;
    va_list argp;
    va_start(argp, fmt);

    tim=time(NULL);
    time_p=localtime(&tim);
    tim_sz=strftime(line, sizeof(line), "%b %d  %H:%M:%S  ", time_p);
    if (log_file!=NULL && log_mode==__LOG_FILE)
    {
	fputs(line, log_file);
	vfprintf(log_file, fmt, argp);
	fputc('\n', log_file);
	return;
    }

    nl=vsnprintf(line+tim_sz, sizeof(line)-tim_sz, fmt, argp)+tim_sz;
    
    newbuf=0;
    bs=bsize;
    while(used+nl+2>=bs)
    {
	if (bs<TEXT_MAX)
	{
	    newbuf=1;
	    bs+=TEXT_INC;
	}
	else
	{
	    if (used>0)
	    {
		/* Save space by throwing existing stuff onto logs */
		if (log_mode!=__LOG_EMAIL || !smtp_send(recipients,  buf))
		{
		    /* Throw it in the logs anyway... */
		    if (buf!=NULL)
			use_syslog(buf);
		    else
			syslog(LOG_WARNING, "%s", line);
		}
		used=0;
	    }
	    else
	    {
		fprintf(stderr, "Fatal: %d+2=nl+2>%d", nl, TEXT_MAX);
		exit(1);	/* Impossible if TEXT_MAX>sizeof(line) */
	    }
	}
    }

    if (newbuf)
    {
	if ((nb=(char *)malloc(bs))==NULL)
	{
	    if (log_mode!=__LOG_EMAIL || !smtp_send(recipients,  buf))
	    {
		/* Throw it in the logs anyway... */
		if (buf!=NULL)
		    use_syslog(buf);
		else
		    syslog(LOG_WARNING, "%s", line);
	    }
	    used=0;
	}
	else
	{
	    memcpy(nb, buf, used);
	    if (buf!=NULL)
		free(buf);
	    buf=nb;
	    bsize=bs;
	}
    }
    memcpy(buf+used, line, nl);
    used+=nl;
    buf[used++]='\n';
    buf[used]='\0';
    va_end(argp);
}

void log_start(void)
{
    if (buf!=NULL)
	log_flush();
    if ((buf=(char *)malloc(INITIAL_BUFFER))==NULL)
    {
	buf=NULL;
	used=0;
	bsize=0;
	return;			/* Not much choice here... */
    }

    bsize=INITIAL_BUFFER;
    used=0;
}

void log_flush(void)
{
    if (used>0)
    {
	if (!smtp_send(recipients, buf))
	    use_syslog(buf);
    }
    if (buf!=NULL)
	free(buf);
    buf=NULL;
    used=0;
    bsize=0;
}

