/* $Id: pidlist_pkill.c,v 1.1.1.1 2000/05/04 22:55:53 dps Exp $ */
/* Read process names from /proc */

#include "config.h"
#include <stdio.h>
#include <ctype.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* #ifdef HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif /* HAVE_DIRENT_H */
#include "cfg_check.h"
#include "prog.h"
#include "numbuf.h"
#include "killscan.h"

/* fs type support */
#ifdef STATFS_WORKS
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif /* HAVE_SYS_VFS_H */
#endif /* STATFS_WORKS */

/* lstat hacks for broken systems without it */
#ifndef HAVE_LSTAT
#define lstat(x,y) stat(x,y)
#endif /* HAVE_LSTAT */

/* Pid tester. Anti-race measure. Also used for extra paranioa in pid_list */
int current_process_pkill(int pid)
{
    if (kill(pid, 0)==-1)
    {
	if (errno==ESRCH)
	    return 0;
	if (errno!=EPERM)
	    log("Unexpected error number %d check pid %d", errno, pid);
    }
    return 1;
}

/* Pid tester. Anti-race measure. Also used for extra paranioa in pid_list */
static int proc_current_process(int pid)
{
    char pathbuf[PROCFNAME_MAX];	/* Lots */
    struct stat stbuf;
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    struct statfs fts;
#endif
    
    if ((snprintf(pathbuf, sizeof(pathbuf), quote(PROCPATH) "/%d",
		  pid))>=sizeof(pathbuf)-1)
    {
	/* I think this is impossible but paranioa is vital here */
	log("Pathbuf overflow pid=%d", pid);
	return 1;			/* Assume process is active */
    }
#if defined(HAVE_STATFS) && defined(CPS_PROC_FS_TYPE) && defined(STATFS_WORKS)
    if (statfs(pathbuf, &fts)!=0)
    {
	if (errno==ENOENT)
	   return 0;
	log("Unexpected failure of statfs (error %d)", errno);
	return 1;
    }
    if (fts.f_type!=CPS_PROC_FS_TYPE)
    {
	log("Uh. Uh. Detected hacking attempt: wrong fs type!");
	return 1;
    }
#endif /* HAVE_STATFS and STATFS_FST and STATFS_WORKS */
    if (lstat(pathbuf, &stbuf)!=0)
    {
	if (errno==ENOENT)
	    return 0;		/* Expected error */
	log("Unexpected errno from lstat: %d", errno); /* Log errno */
	return 1;			/* Assume process is active */
    }
    else
    {
	if ((stbuf.st_mode & S_IFMT)!=CPS_PROC_PIDFMT)
	    log("Hacked libc or /proc. %s wrong type", pathbuf);
	return 1;		/* Process is active */
    }
    /* Not reached */
    log("Serious bug. Code got to current_process IP1");
    return 1;
}

/* build list from numbers listed in /proc */
numbuf *pid_list_pkill(void)
{
    DIR *dp;
    struct dirent *filep;
    int probed_pid;
    numbuf *r;
    pid_t scan;

    if ((r=allocate())==NULL)
    {
	log("Could not allocate space for numbers");
	return NULL;
    }

    for (scan=1; scan!=MAX_PROC; scan++) /* pid_t might be >16 bits wide */
    {
	kill(scan, 0);
	if (errno==ESRCH)
	    continue;
	/* Check for the pid in /proc */
	if (proc_current_process(scan))
	{
	    /* List pid */
	    if (r->pos==r->max && expand_buf(r)==0)
	    {
		free_nums(r);
		return NULL;
	    }
	    r->nums[(r->pos)++]=scan;
	    continue;
	}
	kill(scan, 0);
	if (errno==ESRCH)
	    continue;		/* Just a race */
	log("pid %d active but not in /proc", scan); /* /proc falsifiation */
	nuke_proc(scan);
	kill(scan, 0);
	if (errno!=ESRCH)
	{
	    log("Security Alert! pid %d still active", scan);
	}
    }
    return r;
}
